/* 
 * This is a modified version of...
 *
 * tkListbox.c --
 *
 * Modifications are copyright (c) 1994 Malcolm Beattie.
 * You may distribute under the terms of the GNU General Public License,
 * as specified in the README file.
 *
 *	This module implements listbox widgets for the Tk
 *	toolkit.  A listbox displays a collection of strings,
 *	one per line, and provides scrolling and selection.
 *
 * Copyright (c) 1990-1993 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

/*
 * Multi-selection code ported to tkperl5 by KOBAYASHI Hiroaki
 * <kobayasi@csl1.sowa.is.uec.ac.jp> mostly from Tk 3.x disjoint listbox
 * selection patches written by Marc R. Ewing <meathead@acpub.duke.edu>
 * and Heiko Schroeder <heiko@pool.informatik.rwth-aachen.de>
 */

#ifndef lint
static char rcsid[] = "$Id: tkpListbox.c,v 1.3 1994/10/16 20:01:39 mbeattie Exp $";
#endif

/* $Log: tkpListbox.c,v $
 * Revision 1.3  1994/10/16  20:01:39  mbeattie
 * Multi-selection patch
 *
 * Revision 1.2  1994/10/02  17:26:06  mbeattie
 * curselection method failed to return proper undef when nothing selected.
 *
 * Revision 1.1  1994/09/05  16:03:30  mbeattie
 * Initial revision
 *
 */

#include "tkConfig.h"
#include "default.h"
#include "tkInt.h"
#include "tkperl.h"

WidgetClass listboxClassRec;
I32 listboxMethods[] =
{
    IX_CONFIGURE, IX_CURSELECTION, IX_DELETE, IX_GET, IX_INSERT, IX_NEAREST,
    IX_SCANMARK, IX_SCANDRAGTO, IX_SELECTCLEAR, IX_SELECTADJUST,
    IX_SELECTFROM, IX_SELECTTO, IX_SIZE, IX_XVIEW, IX_YVIEW, 
#ifdef MULTI_SELECTION
    IX_SELECTED, IX_SELECTSET, IX_SELECTINVERT, IX_SELECTTEST,
#endif /* MULTI_SELECTION */
    IX_EOL
};

/*
 * One record of the following type is kept for each element
 * associated with a listbox widget:
 */

typedef struct Element {
#ifdef MULTI_SELECTION
    int selected;               /* 1 if this element is selected */
#endif /* MULTI_SELECTION */
    int textLength;		/* # non-NULL characters in text. */
    int lBearing;		/* Distance from first character's
				 * origin to left edge of character. */
    int pixelWidth;		/* Total width of element in pixels (including
				 * left bearing and right bearing). */
    struct Element *nextPtr;	/* Next in list of all elements of this
				 * listbox, or NULL for last element. */
    char text[4];		/* Characters of this element, NULL-
				 * terminated.  The actual space allocated
				 * here will be as large as needed (> 4,
				 * most likely).  Must be the last field
				 * of the record. */
} Element;

#define ElementSize(stringLength) \
	((unsigned) (sizeof(Element) - 3 + stringLength))

/*
 * A data structure of the following type is kept for each listbox
 * widget managed by this file:
 */

typedef struct {
    Tk_Window tkwin;		/* Window that embodies the listbox.  NULL
				 * means that the window has been destroyed
				 * but the data structures haven't yet been
				 * cleaned up.*/
    Display *display;		/* Display containing widget.  Used, among
				 * other things, so that resources can be
				 * freed even after tkwin has gone away. */
    int numElements;		/* Total number of elements in this listbox. */
    Element *elementPtr;	/* First in list of elements (NULL if no
				 * elements. */

    /*
     * Information used when displaying widget:
     */

    Tk_3DBorder normalBorder;	/* Used for drawing border around whole
				 * window, plus used for background. */
    int borderWidth;		/* Width of 3-D border around window. */
    int relief;			/* 3-D effect: TK_RELIEF_RAISED, etc. */
    XFontStruct *fontPtr;	/* Information about text font, or NULL. */
    XColor *fgColorPtr;		/* Text color in normal mode. */
    GC textGC;			/* For drawing normal text. */
    Tk_3DBorder selBorder;	/* Borders and backgrounds for selected
				 * elements. */
    int selBorderWidth;		/* Width of border around selection. */
    XColor *selFgColorPtr;	/* Foreground color for selected elements. */
    GC selTextGC;		/* For drawing selected text. */
    char *geometry;		/* Desired geometry for window.  Malloc'ed. */
    int lineHeight;		/* Number of pixels allocated for each line
				 * in display. */
    int topIndex;		/* Index of top-most element visible in
				 * window. */
    int numLines;		/* Number of lines (elements) that fit
				 * in window at one time. */
    int setGrid;		/* Non-zero means pass gridding information
				 * to window manager. */

    /*
     * Information to support horizontal scrolling:
     */

    int maxWidth;		/* Width (in pixels) of widest string in
				 * listbox. */
    int xScrollUnit;		/* Number of pixels in one "unit" for
				 * horizontal scrolling (window scrolls
				 * horizontally in increments of this size).
				 * This is an average character size. */
    int xOffset;		/* The left edge of each string in the
				 * listbox is offset to the left by this
				 * many pixels (0 means no offset, positive
				 * means there is an offset). */

    /*
     * Information about what's selected, if any.
     */



#ifdef     MULTI_SELECTION
    int numSelected;            /* Number of elements selected */
    int lastTo;                 /* Index of last select to */
#else  /*  MULTI_SELECTION */
    int selectFirst;		/* Index of first selected element (-1 means
				 * nothing selected. */
    int selectLast;		/* Index of last selected element. */
#endif /*  MULTI_SELECTION */
    int selectAnchor;		/* Fixed end of selection (i.e. element
				 * at which selection was started.) */
    int exportSelection;	/* Non-zero means tie internal listbox
				 * to X selection. */
#ifdef MULTI_SELECTION
    int multipleSelections;     /* Non-zero means only one area of selection
                                 * may exist. (Old behavior) */
#endif /* MULTI_SELECTION */
    /*
     * Information for scanning:
     */

    int scanMarkX;		/* X-position at which scan started (e.g.
				 * button was pressed here). */
    int scanMarkY;		/* Y-position at which scan started (e.g.
				 * button was pressed here). */
    int scanMarkXOffset;	/* Value of "xOffset" field when scan
				 * started. */
    int scanMarkYIndex;		/* Index of line that was at top of window
				 * when scan started. */

    /*
     * Miscellaneous information:
     */

    Cursor cursor;		/* Current cursor for window, or None. */
    struct tkpcallback yScrollCallback;
                                /* Callback slave and method for communicating
				 * with vertical scrollbar.  NULL means no
				 * command to issue. */
    struct tkpcallback xScrollCallback;
                                /* Callback slave and method for communicating
				 * with horizontal scrollbar.  NULL means no
				 * command to issue. */
    int flags;			/* Various flag bits:  see below for
				 * definitions. */
} Listbox;

/*
 * Flag bits for buttons:
 *
 * REDRAW_PENDING:		Non-zero means a DoWhenIdle handler
 *				has already been queued to redraw
 *				this window.
 * UPDATE_V_SCROLLBAR:		Non-zero means vertical scrollbar needs
 *				to be updated.
 * UPDATE_H_SCROLLBAR:		Non-zero means horizontal scrollbar needs
 *				to be updated.
 */

#define REDRAW_PENDING		1
#define UPDATE_V_SCROLLBAR	2
#define UPDATE_H_SCROLLBAR	4

/*
 * Information used for argv parsing:
 */

static Tk_ConfigSpec configSpecs[] = {
    {TK_CONFIG_BORDER, "-background", "background", "Background",
	DEF_LISTBOX_BG_COLOR, Tk_Offset(Listbox, normalBorder),
	TK_CONFIG_COLOR_ONLY},
    {TK_CONFIG_BORDER, "-background", "background", "Background",
	DEF_LISTBOX_BG_MONO, Tk_Offset(Listbox, normalBorder),
	TK_CONFIG_MONO_ONLY},
    {TK_CONFIG_SYNONYM, "-bd", "borderWidth", (char *) NULL,
	(char *) NULL, 0, 0},
    {TK_CONFIG_SYNONYM, "-bg", "background", (char *) NULL,
	(char *) NULL, 0, 0},
    {TK_CONFIG_PIXELS, "-borderwidth", "borderWidth", "BorderWidth",
	DEF_LISTBOX_BORDER_WIDTH, Tk_Offset(Listbox, borderWidth), 0},
    {TK_CONFIG_ACTIVE_CURSOR, "-cursor", "cursor", "Cursor",
	DEF_LISTBOX_CURSOR, Tk_Offset(Listbox, cursor), TK_CONFIG_NULL_OK},
    {TK_CONFIG_BOOLEAN, "-exportselection", "exportSelection",
	"ExportSelection", DEF_LISTBOX_EXPORT_SELECTION,
	Tk_Offset(Listbox, exportSelection), 0},
#ifdef MULTI_SELECTION
    {TK_CONFIG_BOOLEAN, "-multipleselections", "multipleSelections",
        "MultipleSelections", "0",
        Tk_Offset(Listbox, multipleSelections), 0},
#endif /* MULTI_SELECTION */
    {TK_CONFIG_SYNONYM, "-fg", "foreground", (char *) NULL,
	(char *) NULL, 0, 0},
    {TK_CONFIG_FONT, "-font", "font", "Font",
	DEF_LISTBOX_FONT, Tk_Offset(Listbox, fontPtr), 0},
    {TK_CONFIG_COLOR, "-foreground", "foreground", "Foreground",
	DEF_LISTBOX_FG, Tk_Offset(Listbox, fgColorPtr), 0},
    {TK_CONFIG_STRING, "-geometry", "geometry", "Geometry",
	DEF_LISTBOX_GEOMETRY, Tk_Offset(Listbox, geometry), 0},
    {TK_CONFIG_RELIEF, "-relief", "relief", "Relief",
	DEF_LISTBOX_RELIEF, Tk_Offset(Listbox, relief), 0},
    {TK_CONFIG_BORDER, "-selectbackground", "selectBackground", "Foreground",
	DEF_LISTBOX_SELECT_COLOR, Tk_Offset(Listbox, selBorder),
	TK_CONFIG_COLOR_ONLY},
    {TK_CONFIG_BORDER, "-selectbackground", "selectBackground", "Foreground",
	DEF_LISTBOX_SELECT_MONO, Tk_Offset(Listbox, selBorder),
	TK_CONFIG_MONO_ONLY},
    {TK_CONFIG_PIXELS, "-selectborderwidth", "selectBorderWidth", "BorderWidth",
	DEF_LISTBOX_SELECT_BD, Tk_Offset(Listbox, selBorderWidth), 0},
    {TK_CONFIG_COLOR, "-selectforeground", "selectForeground", "Background",
	DEF_LISTBOX_SELECT_FG_COLOR, Tk_Offset(Listbox, selFgColorPtr),
	TK_CONFIG_COLOR_ONLY},
    {TK_CONFIG_COLOR, "-selectforeground", "selectForeground", "Background",
	DEF_LISTBOX_SELECT_FG_MONO, Tk_Offset(Listbox, selFgColorPtr),
	TK_CONFIG_MONO_ONLY},
    {TK_CONFIG_BOOLEAN, "-setgrid", "setGrid", "SetGrid",
	DEF_LISTBOX_SET_GRID, Tk_Offset(Listbox, setGrid), 0},
    {TKP_CONFIG_SLAVE, "-xscrollslave", "xScrollSlave", "ScrollSlave",
	0, Tk_Offset(Listbox, xScrollCallback), TK_CONFIG_NULL_OK},
    {TKP_CONFIG_METHOD, "-xscrollmethod", "xScrollMethod", "ScrollMethod",
	0, Tk_Offset(Listbox, xScrollCallback), TK_CONFIG_NULL_OK},
    {TKP_CONFIG_SLAVE, "-yscrollslave", "yScrollSlave", "ScrollSlave",
	0, Tk_Offset(Listbox, yScrollCallback), TK_CONFIG_NULL_OK},
    {TKP_CONFIG_METHOD, "-yscrollmethod", "yScrollMethod", "ScrollMethod",
	0, Tk_Offset(Listbox, yScrollCallback), TK_CONFIG_NULL_OK},
    {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
	(char *) NULL, 0, 0}
};


/*
 * Forward declarations for procedures defined later in this file:
 */

STATIC void		ChangeListboxOffset _ANSI_ARGS_((Listbox *listPtr,
			    int offset));
STATIC void		ChangeListboxView _ANSI_ARGS_((Listbox *listPtr,
			    int index));
STATIC void		TkpConfigureListbox _ANSI_ARGS_((Listbox *listPtr,
			    int sp, int items, int flags));
STATIC void		DeleteEls _ANSI_ARGS_((Listbox *listPtr, int first,
			    int last));
STATIC void		DestroyListbox _ANSI_ARGS_((ClientData clientData));
STATIC void		DisplayListbox _ANSI_ARGS_((ClientData clientData));
STATIC int		GetListboxIndex _ANSI_ARGS_((Tcl_Interp *interp,
			    Listbox *listPtr, char *string, int endAfter,
			    int *indexPtr));
STATIC void		TkpInsertEls _ANSI_ARGS_((Listbox *listPtr, int index,
			    int sp, int items));
STATIC void		ListboxComputeWidths _ANSI_ARGS_((Listbox *listPtr,
			    int fontChanged));
STATIC void		ListboxEventProc _ANSI_ARGS_((ClientData clientData,
			    XEvent *eventPtr));
STATIC int		ListboxFetchSelection _ANSI_ARGS_((
			    ClientData clientData, int offset, char *buffer,
			    int maxBytes));
STATIC void		ListboxLostSelection _ANSI_ARGS_((
			    ClientData clientData));
STATIC void		ListboxRedrawRange _ANSI_ARGS_((Listbox *listPtr,
			    int first, int last));
STATIC void		ListboxScanTo _ANSI_ARGS_((Listbox *listPtr,
			    int x, int y));
#ifdef MULTI_SELECTION
STATIC void		ListboxSetRange _ANSI_ARGS_((Listbox *listPtr,
			    int index1, int index2, int m));
STATIC void		ListboxUnselectRange _ANSI_ARGS_((Listbox *listPtr,
			    int index1, int index2));
STATIC void		ListboxSelectRange _ANSI_ARGS_((Listbox *listPtr,
			    int index1, int index2));
STATIC void		ListboxInvertRange _ANSI_ARGS_((Listbox *listPtr,
			    int index1, int index2));
STATIC void             ListboxSelectAdjust _ANSI_ARGS_((Listbox *listPtr,
                            int index));
#endif /* MULTI_SELECTION */
STATIC void		ListboxSelectFrom _ANSI_ARGS_((Listbox *listPtr,
			    int index));
STATIC void		ListboxSelectTo _ANSI_ARGS_((Listbox *listPtr,
			    int index));
STATIC void		ListboxUpdateHScrollbar _ANSI_ARGS_((Listbox *listPtr));
STATIC void		ListboxUpdateVScrollbar _ANSI_ARGS_((Listbox *listPtr));
STATIC int		NearestListboxElement _ANSI_ARGS_((Listbox *listPtr,
			    int y));

int
Tk_ListboxCmd(clientData, interp, argc, argv)
ClientData clientData;
Tcl_Interp *interp;
int argc;
char **argv;
{
    croak("shouldn't call Tk_ListboxCmd");
    /* NOTREACHED */
}

/*
 *--------------------------------------------------------------
 *
 * TkpListboxCmd --
 *
 *	This procedure is invoked to process the "listbox"
 *	command.  See the user documentation for details on what
 *	it does.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

I32 TkpListboxCmd(ix, sp, items)
int ix;
int sp;
int items;
{
    SV **st = stack_base + sp;
    register Listbox *listPtr;
    Tk_Window new;
    char *pathname;
    
    if (items < 1)
	croak("wrong # args for new listbox");

    pathname = TkpGenPathname(st[1]);
    new = Tk_CreateWindowFromPath(tclinterp, mainwin, pathname, (char *) NULL);
    if (!new)
	croak("failed to create window");

    /*
     * Initialize the fields of the structure that won't be initialized
     * by ConfigureListbox, or that ConfigureListbox requires to be
     * initialized already (e.g. resource pointers).
     */

    listPtr = (Listbox *) ckalloc(sizeof(Listbox));
    listPtr->tkwin = new;
    listPtr->display = Tk_Display(new);
    listPtr->numElements = 0;
    listPtr->elementPtr = NULL;
    listPtr->normalBorder = NULL;
    listPtr->borderWidth = 0;
    listPtr->relief = TK_RELIEF_RAISED;
    listPtr->fontPtr = NULL;
    listPtr->fgColorPtr = NULL;
    listPtr->textGC = None;
    listPtr->selBorder = NULL;
    listPtr->selBorderWidth = 0;
    listPtr->selFgColorPtr = None;
    listPtr->selTextGC = None;
    listPtr->geometry = NULL;
    listPtr->lineHeight = 0;
    listPtr->topIndex = 0;
    listPtr->numLines = 0;
    listPtr->setGrid = 0;
    listPtr->maxWidth = 0;
    listPtr->xScrollUnit = 0;
    listPtr->xOffset = 0;
#ifdef    MULTI_SELECTION
    listPtr->numSelected = 0;
    listPtr->lastTo = 0;
#else  /* MULTI_SELECTION */
    listPtr->selectFirst = -1;
    listPtr->selectLast = -1;
#endif /* MULTI_SELECTION */
    listPtr->selectAnchor = 0;
    listPtr->exportSelection = 1;
    listPtr->scanMarkX = 0;
    listPtr->scanMarkY = 0;
    listPtr->scanMarkXOffset = 0;
    listPtr->scanMarkYIndex = 0;
    listPtr->cursor = None;
    Zero(&listPtr->xScrollCallback, 1, struct tkpcallback);
    Zero(&listPtr->yScrollCallback, 1, struct tkpcallback);
    listPtr->flags = 0;

    Tk_SetClass(listPtr->tkwin, "Listbox");
    Tk_CreateEventHandler(listPtr->tkwin, ExposureMask|StructureNotifyMask,
	    ListboxEventProc, (ClientData) listPtr);
    Tk_CreateSelHandler(listPtr->tkwin, XA_STRING, ListboxFetchSelection,
	    (ClientData) listPtr, XA_STRING);
    TkpCreateWidgetHash(listPtr->tkwin, (ClientData) listPtr);
    TkpConfigureListbox(listPtr, sp + 1, items - 1, 0);

    /*
     * If the above fails, we ought to Tk_DestroyWindow(listPtr->tkwin)
     * but I've mucked up error propagation. Oops.
     */

    st[0] = sv_2mortal(NewWidgetRef(Tk_PathName(listPtr->tkwin),
				    listboxClassRec.stash));
    return sp;
}

/*
 *--------------------------------------------------------------
 *
 * TkpListboxWidgetCmd --
 *
 *	This procedure is invoked to process the command
 *	that corresponds to a widget managed by this module.
 *	See the user documentation for details on what it does.
 *
 * Side effects:
 *	See the user documentation.
 *
 *--------------------------------------------------------------
 */

I32 TkpListboxWidgetCmd(ix, sp, items)
int ix;
int sp;
int items;
{
    SV **st = stack_base + sp;
    Listbox *listPtr;

    if (items < 1)
	croak("wrong # args");
    if (!WrefToWidget(st[1], 0, (ClientData *) &listPtr, "Listbox", 7))
	croak("non-widget argument passed");

    Tk_Preserve((ClientData) listPtr);

    switch (ix)
    {
      case IX_CONFIGURE:
	if (items == 1)
	    sp = TkpConfigureInfo(sp, listPtr->tkwin, configSpecs,
				  (char *) listPtr, (char *) NULL, 0);
	else if (items == 2)
	    sp = TkpConfigureInfo(sp, listPtr->tkwin, configSpecs,
				  (char *) listPtr, SvPV(st[2], na), 0);
	else
	    TkpConfigureListbox(listPtr, sp + 1, items - 1,
				TK_CONFIG_ARGV_ONLY);
	break;

      case IX_CURSELECTION:
	{
	    int i;

	    if (items != 1)
		croak("wrong # args for curselection");

#ifdef MULTI_SELECTION
	    if (!listPtr->numSelected)
		st[0] = sv_newmortal();
	    else
	    {
		register Element *elPtr = listPtr->elementPtr;

		for (i = 0; i < listPtr->numElements; i++)
		{
		    if (elPtr->selected == 1)
			av_store(stack, sp++, sv_2mortal(newSViv((I32) i)));
		    elPtr = elPtr->nextPtr;
		}
		sp--;
		RECOVERSTACK;
	    }
#else  /* MULTI_SELECTION */
	    if (listPtr->selectFirst == -1)
		st[0] = sv_newmortal();
	    else
	    {
		for (i = listPtr->selectFirst; i <= listPtr->selectLast; i++)
		    av_store(stack, sp++, sv_2mortal(newSViv((I32) i)));
		sp--;
		RECOVERSTACK;
	    }
#endif /* MULTI_SELECTION */
	    break;
	}
      case IX_DELETE:
	{
	    int first, last;

	    if (items < 2 || items > 3)
		croak("wrong # args for delete");
	    if (GetListboxIndex(tclinterp,
				listPtr, SvPV(st[2], na), 0, &first) != TCL_OK)
		croak("bad listbox index");

	    if (items == 2)
		last = first;
	    else
		if (GetListboxIndex(tclinterp,
				    listPtr, SvPV(st[3], na), 0, &last) != TCL_OK)
		croak("bad listbox index");
	    DeleteEls(listPtr, first, last);
	    break;
	}
	
      case IX_GET:
	{
	    int index;
	    register Element *elPtr;

	    if (items != 2)
		croak("wrong # args for get");

	    if (GetListboxIndex(tclinterp,
				listPtr, SvPV(st[2], na), 0, &index) != TCL_OK)
		croak("bad listbox index");

	    if (index < 0)
		index = 0;

	    if (index >= listPtr->numElements)
		index = listPtr->numElements - 1;

	    for (elPtr = listPtr->elementPtr; index > 0;
		 index--, elPtr = elPtr->nextPtr); /* Empty loop body. */

	    if (elPtr)
		st[0] =  sv_2mortal(newSVpv(elPtr->text, 0));
	    else
		st[0] = sv_mortalcopy(&sv_undef);
	    break;
	}

      case IX_INSERT:
	{
	    int index;

	    if (items < 2)
		croak("wrong # args for insert");

	    if (GetListboxIndex(tclinterp,
				listPtr, SvPV(st[2], na), 1, &index) != TCL_OK)
		croak("bad listbox index");


	    TkpInsertEls(listPtr, index, sp + 2, items - 2);
	}
	break;

      case IX_NEAREST:
	if (items != 2)
	    croak("wrong # args for nearest");
	
	st[0] = sv_2mortal(newSViv(NearestListboxElement(listPtr,
							 SvIV(st[2]))));
	                                                  /* Eek */
	break;
	    
#ifdef MULTI_SELECTION
      case IX_SELECTED:
	st[0] = sv_2mortal(newSViv(listPtr->numSelected));
	break;
#endif /* MULTI_SELECTION */

      case IX_SCANMARK:
	if (items != 3)
	    croak("wrong # args for scanmark");
	
	listPtr->scanMarkX = SvIV(st[2]);
	listPtr->scanMarkY = SvIV(st[3]);
	listPtr->scanMarkXOffset = listPtr->xOffset;
	listPtr->scanMarkYIndex = listPtr->topIndex;
	break;
	
      case IX_SCANDRAGTO:
	if (items != 3)
	    croak("wrong # args for scandragto");

	ListboxScanTo(listPtr, SvIV(st[2]), SvIV(st[3]));
	break;

      case IX_SELECTCLEAR:
#ifdef MULTI_SELECTION
	{
	  int x, y;
	  if (items > 3)
	    croak("wrong # args for selectclear");
	  
	  /* argv[3] == SvPV(st[2], na) */
	  if (items == 1) {
	    ListboxUnselectRange(listPtr, 0, listPtr->numElements);
	  } else {
	    if (GetListboxIndex(tclinterp,
				listPtr, SvPV(st[2], na), 0, &x) != TCL_OK) {
	      croak("bad listbox index");
	    } else if (items == 2) {
	      ListboxUnselectRange(listPtr, x, x);
	    } else {
	      if (GetListboxIndex(tclinterp,
				  listPtr, SvPV(st[3], na), 0, &y) != TCL_OK) {
		croak("bad listbox index");
	      } else {
		ListboxUnselectRange(listPtr, x, y);
	      }
	    }
	  }
	}
#else  /* MULTI_SELECTION */
	if (items != 1)
	    croak("wrong # args for selectclear");
	
	if (listPtr->selectFirst != -1) {
	    ListboxRedrawRange(listPtr, listPtr->selectFirst,
			       listPtr->selectLast);
	    listPtr->selectFirst = -1;
	}
#endif /* MULTI_SELECTION */
	break;
#ifdef MULTI_SELECTION
      case IX_SELECTSET:
	{
	  int x, y;
	  if (items > 3)
	    croak("wrong # args for selectclear");
	  
	  if (items == 1) {
	    ListboxSelectRange(listPtr, 0, listPtr->numElements);
	  } else {
	    if (GetListboxIndex(tclinterp,
				listPtr, SvPV(st[2], na), 0, &x) != TCL_OK) {
	      croak("bad listbox index");
	    } else if (items == 2) {
	      ListboxSelectRange(listPtr, x, x);
	    } else {
	      if (GetListboxIndex(tclinterp,
				  listPtr, SvPV(st[3], na), 0, &y) != TCL_OK) {
		croak("bad listbox index");
	      } else {
		ListboxSelectRange(listPtr, x, y);
	      }
	    }
	  }
	  break;
	}
      case IX_SELECTINVERT:
	{
	  int x, y;
	  if (items > 3)
	    croak("wrong # args for Invertclear");
	  
	  if (items == 1) {
	    ListboxInvertRange(listPtr, 0, listPtr->numElements);
	  } else {
	    if (GetListboxIndex(tclinterp,
				listPtr, SvPV(st[2], na), 0, &x) != TCL_OK) {
	      croak("bad listbox index");
	    } else if (items == 2) {
	      ListboxInvertRange(listPtr, x, x);
	    } else {
	      if (GetListboxIndex(tclinterp,
				  listPtr, SvPV(st[3], na), 0, &y) != TCL_OK) {
		croak("bad listbox index");
	      } else {
		ListboxInvertRange(listPtr, x, y);
	      }
	    }
	  }
	  break;
	}
#endif /* MULTI_SELECTION */
      case IX_SELECTADJUST:
	{
	    int index;

	    if (items != 2)
		croak("wrong # args for selectadjust");

	    if (GetListboxIndex(tclinterp,
				listPtr, SvPV(st[2], na), 1, &index) != TCL_OK)
		croak("bad listbox index");

#ifdef     MULTI_SELECTION
	    ListboxSelectAdjust(listPtr, index);
#else  /*  MULTI_SELECTION */
	    if (index < (listPtr->selectFirst + listPtr->selectLast)/2)
		listPtr->selectAnchor = listPtr->selectLast;
	    else
		listPtr->selectAnchor = listPtr->selectFirst;

	    ListboxSelectTo(listPtr, index);
#endif /*  MULTI_SELECTION */
	    break;
	}

#ifdef     MULTI_SELECTION
      case IX_SELECTTEST:
	{
	  int index;
	  register Element *elPtr;
	  if (items != 2)
	    croak("wrong # args for selectfrom");
	
	  if (GetListboxIndex(tclinterp,
			      listPtr, SvPV(st[2], na), 1, &index) != TCL_OK)
	    croak("bad listbox index");

	  elPtr = listPtr->elementPtr;
	  while (index--) {
	    elPtr = elPtr->nextPtr;
	  }
	  st[0] = sv_2mortal(newSViv(elPtr->selected));
	  break;
	}
#endif /*  MULTI_SELECTION */

      case IX_SELECTFROM:
      case IX_SELECTTO:
	{
	    int index;
	    
	    if (items != 2)
		croak("wrong # args for selectfrom");

	    if (GetListboxIndex(tclinterp,
				listPtr, SvPV(st[2], na), 1, &index) != TCL_OK)
		croak("bad listbox index");

	    if (ix == IX_SELECTFROM)
		ListboxSelectFrom(listPtr, index);
	    else
		ListboxSelectTo(listPtr, index);
	    break;
	}

      case IX_SIZE:
	if (items != 1)
	    croak("wrong # args for size");
	st[0] = sv_2mortal(newSViv(listPtr->numElements));
	break;
	
      case IX_XVIEW:
	if (items != 2)
	    croak("wrong # args for xview");

	ChangeListboxOffset(listPtr, SvIV(st[2])*listPtr->xScrollUnit);
	break;

      case IX_YVIEW:
	{
	    int index;

	    if (items != 2)
		croak("wrong # args for yview");

	    if (GetListboxIndex(tclinterp,
				listPtr, SvPV(st[2], na), 0, &index) != TCL_OK)
		croak("bad listbox index");
	    ChangeListboxView(listPtr, index);
	    break;
	}
	
      default:
	croak("bad ix passed");
    }

    RECOVERSTACK;
    Tk_Release((ClientData) listPtr);
    return sp;
}

/*
 *----------------------------------------------------------------------
 *
 * DestroyListbox --
 *
 *	This procedure is invoked by Tk_EventuallyFree or Tk_Release
 *	to clean up the internal structure of a listbox at a safe time
 *	(when no-one is using it anymore).
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Everything associated with the listbox is freed up.
 *
 *----------------------------------------------------------------------
 */

static void
DestroyListbox(clientData)
    ClientData clientData;	/* Info about listbox widget. */
{
    register Listbox *listPtr = (Listbox *) clientData;
    register Element *elPtr, *nextPtr;

    /*
     * Free up all of the list elements.
     */

    for (elPtr = listPtr->elementPtr; elPtr != NULL; ) {
	nextPtr = elPtr->nextPtr;
	ckfree((char *) elPtr);
	elPtr = nextPtr;
    }

    /*
     * Free up all the stuff that requires special handling, then
     * let Tk_FreeOptions handle all the standard option-related
     * stuff.
     */

    if (listPtr->textGC != None) {
	Tk_FreeGC(listPtr->display, listPtr->textGC);
    }
    if (listPtr->selTextGC != None) {
	Tk_FreeGC(listPtr->display, listPtr->selTextGC);
    }
    Tk_FreeOptions(configSpecs, (char *) listPtr, listPtr->display, 0);
    ckfree((char *) listPtr);
}

/*
 *----------------------------------------------------------------------
 *
 * TkpConfigureListbox --
 *
 *	This procedure is called to process the perl stack, plus
 *	the Tk option database, in order to configure (or reconfigure)
 *	a listbox widget.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *	Configuration information, such as colors, border width,
 *	etc. get set for listPtr;  old resources get freed,
 *	if there were any.
 *
 *----------------------------------------------------------------------
 */

static void
TkpConfigureListbox(listPtr, sp, items, flags)
    register Listbox *listPtr;	/* Information about widget;  may or may
				 * not already have values for some fields. */
    int sp;
    int items;
    int flags;			/* Flags to pass to Tk_ConfigureWidget. */
{
    XGCValues gcValues;
    GC new;
    int width, height, fontHeight, oldExport;
    int pixelWidth, pixelHeight;

    oldExport = listPtr->exportSelection;
    TkpConfigureWidget(listPtr->tkwin, configSpecs, sp, items,
		       (char *) listPtr, flags);

    /*
     * A few options need special processing, such as parsing the
     * geometry and setting the background from a 3-D border.
     */

    Tk_SetBackgroundFromBorder(listPtr->tkwin, listPtr->normalBorder);

    gcValues.foreground = listPtr->fgColorPtr->pixel;
    gcValues.font = listPtr->fontPtr->fid;
    gcValues.graphics_exposures = False;
    new = Tk_GetGC(listPtr->tkwin, GCForeground|GCFont|GCGraphicsExposures,
	    &gcValues);
    if (listPtr->textGC != None) {
	Tk_FreeGC(listPtr->display, listPtr->textGC);
    }
    listPtr->textGC = new;

    gcValues.foreground = listPtr->selFgColorPtr->pixel;
    gcValues.font = listPtr->fontPtr->fid;
    new = Tk_GetGC(listPtr->tkwin, GCForeground|GCFont, &gcValues);
    if (listPtr->selTextGC != None) {
	Tk_FreeGC(listPtr->display, listPtr->selTextGC);
    }
    listPtr->selTextGC = new;

    /*
     * Claim the selection if we've suddenly started exporting it.
     */

#ifdef    MULTI_SELECTION
    if (listPtr->exportSelection && (!oldExport) 
	&& (listPtr->numSelected != 0)) 
#else  /* MULTI_SELECTION */
    if (listPtr->exportSelection && (!oldExport) 
	&& (listPtr->selectFirst !=-1)) 
#endif /* MULTI_SELECTION *//* Sorry , I can't keep your indentation policy. */
    {
	Tk_OwnSelection(listPtr->tkwin, ListboxLostSelection,
		(ClientData) listPtr);
    }

    /*
     * Register the desired geometry for the window, and arrange for
     * the window to be redisplayed.
     */

    if ((sscanf(listPtr->geometry, "%dx%d", &width, &height) != 2)
	    || (width <= 0) || (height <= 0))
	croak("bad geometry");

    fontHeight = listPtr->fontPtr->ascent + listPtr->fontPtr->descent;
    listPtr->lineHeight = fontHeight + 1 + 2*listPtr->selBorderWidth;
    listPtr->numLines = (Tk_Height(listPtr->tkwin) - 2*listPtr->borderWidth)
	    / listPtr->lineHeight;
    if (listPtr->numLines < 0) {
	listPtr->numLines = 0;
    }
    ListboxComputeWidths(listPtr, 1);
    pixelWidth = width*listPtr->xScrollUnit + 2*listPtr->borderWidth
	    + 2*listPtr->selBorderWidth;
    pixelHeight = height*listPtr->lineHeight + 2*listPtr->borderWidth;
    Tk_GeometryRequest(listPtr->tkwin, pixelWidth, pixelHeight);
    Tk_SetInternalBorder(listPtr->tkwin, listPtr->borderWidth);
    if (listPtr->setGrid) {
	Tk_SetGrid(listPtr->tkwin, width, height, listPtr->xScrollUnit,
	    listPtr->lineHeight);
    }
    listPtr->flags |= UPDATE_V_SCROLLBAR|UPDATE_H_SCROLLBAR;
    ListboxRedrawRange(listPtr, 0, listPtr->numElements-1);
}

/*
 *--------------------------------------------------------------
 *
 * DisplayListbox --
 *
 *	This procedure redraws the contents of a listbox window.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Information appears on the screen.
 *
 *--------------------------------------------------------------
 */

static void
DisplayListbox(clientData)
    ClientData clientData;	/* Information about window. */
{
    register Listbox *listPtr = (Listbox *) clientData;
    register Tk_Window tkwin = listPtr->tkwin;
    register Element *elPtr;
    GC gc;
    int i, limit, x, y, margin;
    Pixmap pixmap;

    listPtr->flags &= ~REDRAW_PENDING;
    if (listPtr->flags & UPDATE_V_SCROLLBAR) {
	ListboxUpdateVScrollbar(listPtr);
    }
    if (listPtr->flags & UPDATE_H_SCROLLBAR) {
	ListboxUpdateHScrollbar(listPtr);
    }
    listPtr->flags &= ~(REDRAW_PENDING|UPDATE_V_SCROLLBAR|UPDATE_H_SCROLLBAR);
    if ((listPtr->tkwin == NULL) || !Tk_IsMapped(tkwin)) {
	return;
    }

    /*
     * Redrawing is done in a temporary pixmap that is allocated
     * here and freed at the end of the procedure.  All drawing is
     * done to the pixmap, and the pixmap is copied to the screen
     * at the end of the procedure.  This provides the smoothest
     * possible visual effects (no flashing on the screen).
     */

    pixmap = XCreatePixmap(listPtr->display, Tk_WindowId(tkwin),
	    Tk_Width(tkwin), Tk_Height(tkwin), Tk_Depth(tkwin));
    Tk_Fill3DRectangle(listPtr->display, pixmap, listPtr->normalBorder,
	    0, 0, Tk_Width(tkwin), Tk_Height(tkwin), listPtr->borderWidth,
	    listPtr->relief);

    /*
     * Iterate through all of the elements of the listbox, displaying each
     * in turn.  Selected elements use a different GC and have a raised
     * background.
     */

    limit = listPtr->topIndex + listPtr->numLines - 1;
    if (limit >= listPtr->numElements) {
	limit = listPtr->numElements-1;
    }
    margin = listPtr->selBorderWidth + listPtr->xScrollUnit/2;
    for (elPtr = listPtr->elementPtr, i = 0; (elPtr != NULL) && (i <= limit);
	    elPtr = elPtr->nextPtr, i++) {
	if (i < listPtr->topIndex) {
	    continue;
	}
	x = listPtr->borderWidth;
	y = ((i - listPtr->topIndex) * listPtr->lineHeight) 
		+ listPtr->borderWidth;
	gc = listPtr->textGC;
#ifdef    MULTI_SELECTION
	if (elPtr->selected == 1) 
#else   /*  MULTI_SELECTION */
	if ((listPtr->selectFirst >= 0) && (i >= listPtr->selectFirst)
		&& (i <= listPtr->selectLast))
#endif  /*  MULTI_SELECTION */
	{
	    gc = listPtr->selTextGC;
	    Tk_Fill3DRectangle(listPtr->display, pixmap,
		    listPtr->selBorder, x, y,
		    Tk_Width(tkwin) - 2*listPtr->borderWidth,
		    listPtr->lineHeight, listPtr->selBorderWidth,
		    TK_RELIEF_RAISED);
	}
	y += listPtr->fontPtr->ascent + listPtr->selBorderWidth;
	x += margin - elPtr->lBearing - listPtr->xOffset;
	XDrawString(listPtr->display, pixmap, gc, x, y,
		elPtr->text, elPtr->textLength);
    }

    /*
     * Redraw the border for the listbox to make sure that it's on top
     * of any of the text of the listbox entries.
     */

    Tk_Draw3DRectangle(listPtr->display, pixmap,
	    listPtr->normalBorder, 0, 0, Tk_Width(tkwin),
	    Tk_Height(tkwin), listPtr->borderWidth,
	    listPtr->relief);
    XCopyArea(listPtr->display, pixmap, Tk_WindowId(tkwin),
	    listPtr->textGC, 0, 0, Tk_Width(tkwin), Tk_Height(tkwin),
	    0, 0);
    XFreePixmap(listPtr->display, pixmap);
}

/*
 *----------------------------------------------------------------------
 *
 * TkpInsertEls --
 *
 *	Add new elements to a listbox widget.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	New information gets added to listPtr;  it will be redisplayed
 *	soon, but not immediately.
 *
 *----------------------------------------------------------------------
 */

static void
TkpInsertEls(listPtr, index, sp, items)
    register Listbox *listPtr;	/* Listbox that is to get the new
				 * elements. */
    int index;			/* Add the new elements before this
				 * element. */
    int sp;
    int items;
{
    SV **st = stack_base + sp;
    register Element *prevPtr, *newPtr;
    int length, dummy, i, oldMaxWidth;
    XCharStruct bbox;
    char *newel;
#ifdef     MULTI_SELECTION
    int selected;
#endif /*  MULTI_SELECTION */
    

    /*
     * Find the element before which the new ones will be inserted.
     */

    if (index <= 0) {
	index = 0;
    }
    if (index > listPtr->numElements) {
	index = listPtr->numElements;
    }
    if (index == 0) {
	prevPtr = NULL;
    } else {
	for (prevPtr = listPtr->elementPtr, i = index - 1; i > 0; i--) {
	    prevPtr = prevPtr->nextPtr;
	}
    }

#ifdef     MULTI_SELECTION
    selected = (prevPtr) && (prevPtr->selected) && 
	       (prevPtr->nextPtr) && (prevPtr->nextPtr->selected);
#endif /*  MULTI_SELECTION */

    /*
     * For each new element, create a record, initialize it, and link
     * it into the list of elements.
     */

    oldMaxWidth = listPtr->maxWidth;
    for (i = 1; i <= items; i++, prevPtr = newPtr) {
	newel = SvPV(st[i], na);
	length = strlen(newel);
	newPtr = (Element *) ckalloc(ElementSize(length));
#ifdef     MULTI_SELECTION
	newPtr->selected = selected;
#endif /*  MULTI_SELECTION */
	newPtr->textLength = length;
	strcpy(newPtr->text, newel);
	XTextExtents(listPtr->fontPtr, newPtr->text, newPtr->textLength,
		&dummy, &dummy, &dummy, &bbox);
	newPtr->lBearing = bbox.lbearing;
	newPtr->pixelWidth = bbox.rbearing - bbox.lbearing;
	if (newPtr->pixelWidth > listPtr->maxWidth) {
	    listPtr->maxWidth = newPtr->pixelWidth;
	}
	if (prevPtr == NULL) {
	    newPtr->nextPtr = listPtr->elementPtr;
	    listPtr->elementPtr = newPtr;
	} else {
	    newPtr->nextPtr = prevPtr->nextPtr;
	    prevPtr->nextPtr = newPtr;
	}
    }
    listPtr->numElements += items;

    /*
     * Arrange for the new information to be displayed.
     */

#ifndef    MULTI_SELECTION
    if (index <= listPtr->selectFirst) {
	listPtr->selectFirst += items;
    }
    if (index <= listPtr->selectLast) {
	listPtr->selectLast += items;
    }
#endif  /* ! MULTI_SELECTION */
    listPtr->flags |= UPDATE_V_SCROLLBAR;
    if (listPtr->maxWidth != oldMaxWidth) {
	listPtr->flags |= UPDATE_H_SCROLLBAR;
    }
    ListboxRedrawRange(listPtr, index, listPtr->numElements-1);
}

/*
 *----------------------------------------------------------------------
 *
 * DeleteEls --
 *
 *	Remove one or more elements from a listbox widget.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Memory gets freed, the listbox gets modified and (eventually)
 *	redisplayed.
 *
 *----------------------------------------------------------------------
 */

static void
DeleteEls(listPtr, first, last)
    register Listbox *listPtr;	/* Listbox widget to modify. */
    int first;			/* Index of first element to delete. */
    int last;			/* Index of last element to delete. */
{
    register Element *prevPtr, *elPtr;
    int count, i, widthChanged;

    /*
     * Adjust the range to fit within the existing elements of the
     * listbox, and make sure there's something to delete.
     */

#ifdef    MULTI_SELECTION
    if (first <= 0) {
	first = 0;
    } else if (first >= listPtr->numElements) {
	first = listPtr->numElements-1;
    }
#else   /* MULTI_SELECTION */
    if (first < 0) {
	first = 0;
    }
#endif  /* MULTI_SELECTION */
    if (last >= listPtr->numElements) {
	last = listPtr->numElements-1;
    }
    count = last + 1 - first;
    if (count <= 0) {
	return;
    }

    /*
     * Find the element just before the ones to delete.
     */

    if (first == 0) {
	prevPtr = NULL;
    } else {
	for (i = first-1, prevPtr = listPtr->elementPtr; i > 0; i--) {
	    prevPtr = prevPtr->nextPtr;
	}
    }

    /*
     * Delete the requested number of elements.
     */

    widthChanged = 0;
    for (i = count; i > 0; i--) {
	if (prevPtr == NULL) {
	    elPtr = listPtr->elementPtr;
	    listPtr->elementPtr = elPtr->nextPtr;
	} else {
	    elPtr = prevPtr->nextPtr;
	    prevPtr->nextPtr = elPtr->nextPtr;
	}
	if (elPtr->pixelWidth == listPtr->maxWidth) {
	    widthChanged = 1;
	}
#ifdef     MULTI_SELECTION
	listPtr->numSelected -= elPtr->selected;
#endif /*  MULTI_SELECTION */
	ckfree((char *) elPtr);
    }
    listPtr->numElements -= count;

    /*
     * Update the viewing information to reflect the change
     * in the element numbering, and redisplay to slide information up over
     * the elements that were deleted.
     */

#ifndef     MULTI_SELECTION
    if (first <= listPtr->selectFirst) {
	listPtr->selectFirst -= count;
	if (listPtr->selectFirst < first) {
	    listPtr->selectFirst = first;
	}
    }
    if (first <= listPtr->selectLast) {
	listPtr->selectLast -= count;
	if (listPtr->selectLast < first) {
	    listPtr->selectLast = first-1;
	}
    }
    if (listPtr->selectLast < listPtr->selectFirst) {
	listPtr->selectFirst = -1;
    }
#endif /*  ! MULTI_SELECTION */
    if (first <= listPtr->topIndex) {
	listPtr->topIndex -= count;
	if (listPtr->topIndex < first) {
	    listPtr->topIndex = first;
	}
    }
    listPtr->flags |= UPDATE_V_SCROLLBAR;
    if (widthChanged) {
	ListboxComputeWidths(listPtr, 0);
	listPtr->flags |= UPDATE_H_SCROLLBAR;
    }
    ListboxRedrawRange(listPtr, first, listPtr->numElements-1);
}

/*
 *--------------------------------------------------------------
 *
 * ListboxEventProc --
 *
 *	This procedure is invoked by the Tk dispatcher for various
 *	events on listboxes.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	When the window gets deleted, internal structures get
 *	cleaned up.  When it gets exposed, it is redisplayed.
 *
 *--------------------------------------------------------------
 */

static void
ListboxEventProc(clientData, eventPtr)
    ClientData clientData;	/* Information about window. */
    XEvent *eventPtr;		/* Information about event. */
{
    Listbox *listPtr = (Listbox *) clientData;

    if (eventPtr->type == Expose) {
	ListboxRedrawRange(listPtr,
		NearestListboxElement(listPtr, eventPtr->xexpose.y),
		NearestListboxElement(listPtr, eventPtr->xexpose.y
		+ eventPtr->xexpose.height));
    } else if (eventPtr->type == DestroyNotify) {
	TkpDeleteWidgetHash(listPtr->tkwin);
	listPtr->tkwin = NULL;
	if (listPtr->flags & REDRAW_PENDING) {
	    Tk_CancelIdleCall(DisplayListbox, (ClientData) listPtr);
	}
	Tk_EventuallyFree((ClientData) listPtr, DestroyListbox);
    } else if (eventPtr->type == ConfigureNotify) {
	Tk_Preserve((ClientData) listPtr);
	listPtr->numLines = (Tk_Height(listPtr->tkwin)
		- 2*listPtr->borderWidth) / listPtr->lineHeight;
	listPtr->flags |= UPDATE_V_SCROLLBAR|UPDATE_H_SCROLLBAR;
	ListboxRedrawRange(listPtr, 0, listPtr->numElements-1);
	Tk_Release((ClientData) listPtr);
    }
}

/*
 *--------------------------------------------------------------
 *
 * GetListboxIndex --
 *
 *	Parse an index into a listbox and return either its value
 *	or an error.
 *
 * Results:
 *	A standard Tcl result.  If all went well, then *indexPtr is
 *	filled in with the index (into listPtr) corresponding to
 *	string.  Otherwise an error message is left in interp->result.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

static int
GetListboxIndex(interp, listPtr, string, endAfter, indexPtr)
    Tcl_Interp *interp;		/* For error messages. */
    Listbox *listPtr;		/* Listbox for which the index is being
				 * specified. */
    char *string;		/* Numerical index into listPtr's element
				 * list, or "end" to refer to last element. */
    int endAfter;		/* 0 means "end" refers to the index of the
				 * last element, 1 means it refers to the
				 * element after the last one. */
    int *indexPtr;		/* Where to store converted index. */
{
    if (string[0] == 'e') {
	if (strncmp(string, "end", strlen(string)) != 0) {
	    badIndex:
	    Tcl_AppendResult(interp, "bad listbox index \"", string,
		    "\"", (char *) NULL);
	    return TCL_ERROR;
	}
	*indexPtr = listPtr->numElements;
	if (!endAfter) {
	    *indexPtr -= 1;
	}
	if (listPtr->numElements <= 0) {
	    *indexPtr = 0;
	}
    } else {
	if (Tcl_GetInt(interp, string, indexPtr) != TCL_OK) {
	    Tcl_ResetResult(interp);
	    goto badIndex;
	}
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * ChangeListboxView --
 *
 *	Change the view on a listbox widget.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	What's displayed on the screen is changed.  If there is a
 *	scrollbar associated with this widget, then the scrollbar
 *	is instructed to change its display too.
 *
 *----------------------------------------------------------------------
 */

static void
ChangeListboxView(listPtr, index)
    register Listbox *listPtr;		/* Information about widget. */
    int index;				/* Index of element in listPtr. */
{
    if (index >= listPtr->numElements) {
	index = listPtr->numElements-1;
    }
    if (index < 0) {
	index = 0;
    }
    if (listPtr->topIndex != index) {
	listPtr->topIndex = index;
	if (!(listPtr->flags & REDRAW_PENDING)) {
	    Tk_DoWhenIdle(DisplayListbox, (ClientData) listPtr);
	    listPtr->flags |= REDRAW_PENDING;
	}
	listPtr->flags |= UPDATE_V_SCROLLBAR;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * ChangListboxOffset --
 *
 *	Change the horizontal offset for a listbox.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The listbox may be redrawn to reflect its new horizontal
 *	offset.
 *
 *----------------------------------------------------------------------
 */

static void
ChangeListboxOffset(listPtr, offset)
    register Listbox *listPtr;		/* Information about widget. */
    int offset;				/* Desired new "xOffset" for
					 * listbox. */
{
    int maxOffset;

    /*
     * Make sure that the new offset is within the allowable range, and
     * round it off to an even multiple of xScrollUnit.
     */

    maxOffset = listPtr->maxWidth + (listPtr->xScrollUnit-1)
	    - (Tk_Width(listPtr->tkwin) - 2*listPtr->borderWidth
	    - 2*listPtr->selBorderWidth - listPtr->xScrollUnit);
    if (offset > maxOffset) {
	offset = maxOffset;
    }
    if (offset < 0) {
	offset = 0;
    }
    offset -= offset%listPtr->xScrollUnit;
    if (offset != listPtr->xOffset) {
	listPtr->xOffset = offset;
	listPtr->flags |= UPDATE_H_SCROLLBAR;
	ListboxRedrawRange(listPtr, 0, listPtr->numElements);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * ListboxScanTo --
 *
 *	Given a point (presumably of the curent mouse location)
 *	drag the view in the window to implement the scan operation.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The view in the window may change.
 *
 *----------------------------------------------------------------------
 */

static void
ListboxScanTo(listPtr, x, y)
    register Listbox *listPtr;		/* Information about widget. */
    int x;				/* X-coordinate to use for scan
					 * operation. */
    int y;				/* Y-coordinate to use for scan
					 * operation. */
{
    int newTopIndex, newOffset;

    /*
     * Compute new top line for screen by amplifying the difference
     * between the current position and the place where the scan
     * started (the "mark" position).  If we run off the top or bottom
     * of the list, then reset the mark point so that the current
     * position continues to correspond to the edge of the window.
     * This means that the picture will start dragging as soon as the
     * mouse reverses direction (without this reset, might have to slide
     * mouse a long ways back before the picture starts moving again).
     */

    newTopIndex = listPtr->scanMarkYIndex
	    - (10*(y - listPtr->scanMarkY))/listPtr->lineHeight;
    if (newTopIndex >= listPtr->numElements) {
	newTopIndex = listPtr->scanMarkYIndex = listPtr->numElements-1;
	listPtr->scanMarkY = y;
    } else if (newTopIndex < 0) {
	newTopIndex = listPtr->scanMarkYIndex = 0;
	listPtr->scanMarkY = y;
    }
    ChangeListboxView(listPtr, newTopIndex);

    /*
     * Compute new left edge for display in a similar fashion by amplifying
     * the difference between the current position and the place where the
     * scan started.
     */

    newOffset = listPtr->scanMarkXOffset - (10*(x - listPtr->scanMarkX));
    if (newOffset >= listPtr->maxWidth) {
	newOffset = listPtr->scanMarkXOffset = listPtr->maxWidth;
	listPtr->scanMarkX = x;
    } else if (newOffset < 0) {
	newOffset = listPtr->scanMarkXOffset = 0;
	listPtr->scanMarkX = x;
    }
    ChangeListboxOffset(listPtr, newOffset);
}

/*
 *----------------------------------------------------------------------
 *
 * NearestListboxElement --
 *
 *	Given a y-coordinate inside a listbox, compute the index of
 *	the element under that y-coordinate (or closest to that
 *	y-coordinate).
 *
 * Results:
 *	The return value is an index of an element of listPtr.  If
 *	listPtr has no elements, then 0 is always returned.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
NearestListboxElement(listPtr, y)
    register Listbox *listPtr;		/* Information about widget. */
    int y;				/* Y-coordinate in listPtr's window. */
{
    int index;

    index = (y - listPtr->borderWidth)/listPtr->lineHeight;
    if (index >= listPtr->numLines) {
	index = listPtr->numLines-1;
    }
    if (index < 0) {
	index = 0;
    }
    index += listPtr->topIndex;
    if (index >= listPtr->numElements) {
	index = listPtr->numElements-1;
    }
    return index;
}

/*
 *----------------------------------------------------------------------
 *
 * ListboxSelectAdjust --
 *
 *      Adjust the selection
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      ListPtr claims the selection.  The selection is adjusted
 *      to include elements between index and the selected element
 *      closest to index.  If there is no selection, elements 0
 *      through index are selected.
 *
 *----------------------------------------------------------------------
 */

static void
ListboxSelectAdjust(listPtr, index)
    Listbox *listPtr;
    int index;
{
    int up, down, nearest, i;
    Element *elPtr, *uPtr, *iPtr;

    /*
     * Make sure the index is within the proper range for the listbox.
     */

    if (listPtr->numElements == 0) {
	return;
    }
    if (index >= listPtr->numElements) {
	index = listPtr->numElements-1;
    }
    if (index < 0) {
	index = 0;
    }
    
    if ((listPtr->numSelected == 0) && (listPtr->exportSelection)) {
	Tk_OwnSelection(listPtr->tkwin, ListboxLostSelection,
		(ClientData) listPtr);
    }

    if (listPtr->numSelected == 0) {
	nearest = 0;
	elPtr = listPtr->elementPtr;
	i = index + 1;
	goto skip;
    }

    elPtr = listPtr->elementPtr;
    i = 0;
    up = down = -1;
    uPtr = iPtr = NULL;
    while (elPtr != NULL) {
	if (i == index) {
	    iPtr = elPtr;
	}
	if (elPtr->selected) {
	    if (i <= index) {
		uPtr = elPtr;
		up = i;
	    } else {
		down = i;
		break;
	    }
	}
	i++;
	elPtr = elPtr->nextPtr;
    }

    if (up == -1) {
	if (down == -1) {
	    nearest = 0;
	    elPtr = listPtr->elementPtr;
	    i = index + 1;
	} else {
	    nearest = down;
	    elPtr = iPtr;
	    i = down - index + 1;
	}
    } else {
	if (down == -1) {
	    nearest = up;
	    elPtr = uPtr;
	    i = index - up + 1;
	} else {
	    if ((index - up) <= (down - index)) {
		nearest = up;
		elPtr = uPtr;
		i = index - up + 1;
	    } else {
		nearest = down;
		elPtr = iPtr;
		i = down - index + 1;
	    }
	}
    }

  skip:

    while (i) {
	listPtr->numSelected += 1 - elPtr->selected;
	elPtr->selected = 1;
	elPtr = elPtr->nextPtr;
	i--;
    }

    listPtr->selectAnchor = nearest;
    listPtr->lastTo = index;
    ListboxRedrawRange(listPtr, nearest, index);
}

/*
 *----------------------------------------------------------------------
 *
 * ListboxSelectFrom --
 *
 *	Start a new selection in a listbox.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	ListPtr claims the selection, and a selection starts at index.
 *
 *----------------------------------------------------------------------
 */

static void
ListboxSelectFrom(listPtr, index)
    register Listbox *listPtr;		/* Information about widget. */
    int index;				/* Index of element that is to
					 * become the new selection. */
{
#ifdef     MULTI_SELECTION
    int i;
    Element *elPtr;
#endif /*  MULTI_SELECTION */

    /*
     * Make sure the index is within the proper range for the listbox.
     */

#ifdef     MULTI_SELECTION
    if (listPtr->numElements == 0) {
 	return;
    }
#endif /*  MULTI_SELECTION */

    if (index <= 0) {
	index = 0;
    }
    if (index >= listPtr->numElements) {
	index = listPtr->numElements-1;
    }

#ifdef     MULTI_SELECTION
    if ((listPtr->numSelected == 0) && (listPtr->exportSelection)) {
	Tk_OwnSelection(listPtr->tkwin, ListboxLostSelection,
		(ClientData) listPtr);
    }
    /* Allow only one selected area in "compatibility mode" */
    if (!listPtr->multipleSelections && listPtr->numSelected > 0) {
       ListboxUnselectRange(listPtr, 0, listPtr->numElements);
    }

    elPtr = listPtr->elementPtr;
    i = index;
    while (i) {
	elPtr = elPtr->nextPtr;
	i--;
    }

    listPtr->numSelected += 1 - elPtr->selected;
    elPtr->selected = 1;
    listPtr->lastTo = index;
#else  /*  MULTI_SELECTION */
    if (listPtr->selectFirst != -1) {
	ListboxRedrawRange(listPtr, listPtr->selectFirst, listPtr->selectLast);
    } else if (listPtr->exportSelection) {
	Tk_OwnSelection(listPtr->tkwin, ListboxLostSelection,
		(ClientData) listPtr);
    }

    listPtr->selectFirst = listPtr->selectLast = index;
#endif /*  MULTI_SELECTION */
    listPtr->selectAnchor = index;
    ListboxRedrawRange(listPtr, index, index);
}

/*
 *----------------------------------------------------------------------
 *
 * ListboxSelectTo --
 *
 *	Modify the selection by moving its un-anchored end.  This could
 *	make the selection either larger or smaller.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The selection changes.
 *
 *----------------------------------------------------------------------
 */

static void
ListboxSelectTo(listPtr, index)
    register Listbox *listPtr;		/* Information about widget. */
    int index;				/* Index of element that is to
					 * become the "other" end of the
					 * selection. */
{
    register Element *elPtr;
    int i, count;

    /*
     * Make sure the index is within the proper range for the listbox.
     */

    if (listPtr->numElements == 0) {
	return;
    }
    if (index >= listPtr->numElements) {
	index = listPtr->numElements-1;
    }
    if (index < 0) {
	index = 0;
    }

    /*
     * We should already own the selection, but grab it if we don't.
     */

    if (listPtr->numSelected == 0) {
	ListboxSelectFrom(listPtr, index);
	goto done;
    }

    if (index > listPtr->selectAnchor) {
	i = listPtr->selectAnchor;
	count = index - i + 1;
    } else {
	i = index;
	count = listPtr->selectAnchor - index + 1;
    }
    elPtr = listPtr->elementPtr;
    while (i) {
	elPtr = elPtr->nextPtr;
	i--;
    }

    ListboxUnselectRange(listPtr, listPtr->selectAnchor, listPtr->lastTo);

    while (count) {
	listPtr->numSelected += 1 - elPtr->selected;
	elPtr->selected = 1;
	elPtr = elPtr->nextPtr;
	count--;
    }

    listPtr->lastTo = index;
  done:
    ListboxRedrawRange(listPtr, listPtr->selectAnchor, index);
}

/*
 *----------------------------------------------------------------------
 *
 * ListboxFetchSelection --
 *
 *	This procedure is called back by Tk when the selection is
 *	requested by someone.  It returns part or all of the selection
 *	in a buffer provided by the caller.
 *
 * Results:
 *	The return value is the number of non-NULL bytes stored
 *	at buffer.  Buffer is filled (or partially filled) with a
 *	NULL-terminated string containing part or all of the selection,
 *	as given by offset and maxBytes.  The selection is returned
 *	as a Tcl list with one list element for each element in the
 *	listbox.
 *
 * Side effects:
 *	None.
 *
 *----------------------------------------------------------------------
 */

static int
ListboxFetchSelection(clientData, offset, buffer, maxBytes)
    ClientData clientData;		/* Information about listbox widget. */
    int offset;				/* Offset within selection of first
					 * byte to be returned. */
    char *buffer;			/* Location in which to place
					 * selection. */
    int maxBytes;			/* Maximum number of bytes to place
					 * at buffer, not including terminating
					 * NULL character. */
{
    register Listbox *listPtr = (Listbox *) clientData;
    register Element *elPtr;
    char **argv, *selection;
    int src, dst, length, count, argc;

#ifdef     MULTI_SELECTION
    if ((listPtr->numSelected == 0) || !listPtr->exportSelection) 
#else  /*  MULTI_SELECTION */
    if ((listPtr->selectFirst == -1) || !listPtr->exportSelection) 
#endif /*  MULTI_SELECTION */
    {
	return -1;
    }

    /*
     * Use Tcl_Merge to format the listbox elements into a suitable
     * Tcl list.
     */

#ifdef     MULTI_SELECTION
    count = listPtr->numElements;
    argv = (char **) ckalloc((unsigned) (count*sizeof(char *)));
    for (argc = 0, elPtr = listPtr->elementPtr; count > 0;
	 count--, elPtr = elPtr->nextPtr) {
      if (elPtr->selected == 1) {
	argv[argc] = elPtr->text;
      }
    }
#else  /*  MULTI_SELECTION */
    argc = listPtr->selectLast - listPtr->selectFirst + 1;
    argv = (char **) ckalloc((unsigned) (argc*sizeof(char *)));
    for (src = 0, dst = 0, elPtr = listPtr->elementPtr; ;
	    src++, elPtr = elPtr->nextPtr) {
	if (src < listPtr->selectFirst) {
	    continue;
	}
	if (src > listPtr->selectLast) {
	    break;
	}
	argv[dst] = elPtr->text;
	dst++;
    }
#endif /*  MULTI_SELECTION */
    selection = Tcl_Merge(argc, argv);

    /*
     * Copy the requested portion of the selection to the buffer.
     */

    length = strlen(selection);
    count = length - offset;
    if (count <= 0) {
	count = 0;
	goto done;
    }
    if (count > maxBytes) {
	count = maxBytes;
    }
    memcpy((VOID *) buffer, (VOID *) (selection + offset), count);

    done:
    buffer[count] = '\0';
    ckfree(selection);
    ckfree((char *) argv);
    return count;
}

/*
 *----------------------------------------------------------------------
 *
 * ListboxSelectRange
 *
 *      Select a range of elements.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The range of elements is selected.
 *
 *----------------------------------------------------------------------
 */

static void
ListboxSelectRange(listPtr, index1, index2)
    Listbox *listPtr;
    int index1, index2;
{
    ListboxSetRange(listPtr, index1, index2, 1);
}

/*
 *----------------------------------------------------------------------
 *
 * ListboxUnselectRange
 *
 *      Unselect a range of elements.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The range of elements is unselected.
 *
 *----------------------------------------------------------------------
 */

static void
ListboxUnselectRange(listPtr, index1, index2)
    Listbox *listPtr;
    int index1, index2;
{
    ListboxSetRange(listPtr, index1, index2, 0);
}

/*
 *----------------------------------------------------------------------
 *
 * ListboxInvertRange
 *
 *      Invert the selected field of a range of elements.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The range of elements is selected/unselected.
 *
 *----------------------------------------------------------------------
 */

static void
ListboxInvertRange(listPtr, index1, index2)
    Listbox *listPtr;
    int index1, index2;
{
    ListboxSetRange(listPtr, index1, index2, 2);
}

/*
 *----------------------------------------------------------------------
 *
 * ListboxSetRange
 *
 *      Set the selected field of a range of elements.
 *      m=2 means invert the fields.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The range of elements is (un)selected.
 *
 *----------------------------------------------------------------------
 */

static void
ListboxSetRange(listPtr, index1, index2, m)
    Listbox *listPtr;
    int index1, index2, m;
{
    int count, x;
    Element *elPtr;

    /*
     * Make sure the indices are within the proper range for the listbox.
     */

    if (listPtr->numElements == 0) {
	return;
    }
    if (index1 >= listPtr->numElements) {
	index1 = listPtr->numElements-1;
    }
    if (index2 >= listPtr->numElements) {
	index2 = listPtr->numElements-1;
    }
    if (index1 < 0) {
	index1 = 0;
    }
    if (index2 < 0) {
	index2 = 0;
    }
    
    if (index1 > index2) {
	count = index1;
	index1 = index2;
	index2 = count;
    }

    elPtr = listPtr->elementPtr;
    count = 0;
    while (count <= index2) {
	if (count >= index1) {
	    x = elPtr->selected;
	    elPtr->selected = (m == 2) ? (!(elPtr->selected)) : m;
	    listPtr->numSelected += elPtr->selected - x;
	}
	elPtr = elPtr->nextPtr;
	count++;
    }
    ListboxRedrawRange(listPtr, index1, index2);
}

/*
 *----------------------------------------------------------------------
 *
 * ListboxLostSelection --
 *
 *	This procedure is called back by Tk when the selection is
 *	grabbed away from a listbox widget.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The existing selection is unhighlighted, and the window is
 *	marked as not containing a selection.
 *
 *----------------------------------------------------------------------
 */

static void
ListboxLostSelection(clientData)
    ClientData clientData;		/* Information about listbox widget. */
{
    register Listbox *listPtr = (Listbox *) clientData;

#ifdef     MULTI_SELECTION
    if (listPtr->exportSelection) {
	ListboxUnselectRange(listPtr, 0, listPtr->numElements);
    }
#else    /* MULTI_SELECTION */
    if ((listPtr->selectFirst >= 0) && listPtr->exportSelection) {
	ListboxRedrawRange(listPtr, listPtr->selectFirst, listPtr->selectLast);
	listPtr->selectFirst = -1;
    }
#endif   /* MULTI_SELECTION */
}

/*
 *----------------------------------------------------------------------
 *
 * ListboxRedrawRange --
 *
 *	Ensure that a given range of elements is eventually redrawn on
 *	the display (if those elements in fact appear on the display).
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Information gets redisplayed.
 *
 *----------------------------------------------------------------------
 */

	/* ARGSUSED */
static void
ListboxRedrawRange(listPtr, first, last)
    register Listbox *listPtr;		/* Information about widget. */
    int first;				/* Index of first element in list
					 * that needs to be redrawn. */
    int last;				/* Index of last element in list
					 * that needs to be redrawn.  May
					 * be less than first;
					 * these just bracket a range. */
{
    if ((listPtr->tkwin == NULL) || !Tk_IsMapped(listPtr->tkwin)
	    || (listPtr->flags & REDRAW_PENDING)) {
	return;
    }
    Tk_DoWhenIdle(DisplayListbox, (ClientData) listPtr);
    listPtr->flags |= REDRAW_PENDING;
}

/*
 *----------------------------------------------------------------------
 *
 * TkpListboxUpdateVScrollbar --
 *
 *	This procedure is invoked whenever information has changed in
 *	a listbox in a way that would invalidate a vertical scrollbar
 *	display.  If there is an associated scrollbar, then this command
 *	updates it by invoking a perl method in it.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	A perl command is invoked.
 *
 *----------------------------------------------------------------------
 */

static void
ListboxUpdateVScrollbar(listPtr)
    register Listbox *listPtr;		/* Information about widget. */
{
    int last;
    int sp = stack_sp - stack_base;

    if (!CALLBACKOK(&listPtr->yScrollCallback))
	return;

    last = listPtr->topIndex + listPtr->numLines - 1;
    if (last >= listPtr->numElements) {
	last = listPtr->numElements-1;
    }
    if (last < listPtr->topIndex) {
	last = listPtr->topIndex;
    }
    av_store(stack, sp + 2, sv_2mortal(newSViv(listPtr->numElements)));
    av_store(stack, sp + 3, sv_2mortal(newSViv(listPtr->numLines)));
    av_store(stack, sp + 4, sv_2mortal(newSViv(listPtr->topIndex)));
    av_store(stack, sp + 5, sv_2mortal(newSViv(last)));
    RECOVERSTACK;
    TkpCallback(&listPtr->yScrollCallback, 4);
}

/*
 *----------------------------------------------------------------------
 *
 * ListboxUpdateHScrollbar --
 *
 *	This procedure is invoked whenever information has changed in
 *	a listbox in a way that would invalidate a horizontal scrollbar
 *	display.  If there is an associated horizontal scrollbar, then
 *	this command updates it by invoking a Tcl command.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	A Tcl command is invoked, and an additional command may be
 *	invoked to process errors in the command.
 *
 *----------------------------------------------------------------------
 */

static void
ListboxUpdateHScrollbar(listPtr)
    register Listbox *listPtr;		/* Information about widget. */
{
    int totalUnits, windowUnits, first, last;
    int sp = stack_sp - stack_base;

    if (!CALLBACKOK(&listPtr->xScrollCallback))
	return;

    totalUnits = 1 + (listPtr->maxWidth-1)/listPtr->xScrollUnit;
    windowUnits = 1 + (Tk_Width(listPtr->tkwin)
	    - 2*(listPtr->borderWidth + listPtr->selBorderWidth)-1)
	    /listPtr->xScrollUnit;
    first = listPtr->xOffset/listPtr->xScrollUnit;
    last = first + windowUnits - 1;
    if (last < first) {
	last = first;
    }
    av_store(stack, sp + 2, sv_2mortal(newSViv(totalUnits)));
    av_store(stack, sp + 3, sv_2mortal(newSViv(windowUnits)));
    av_store(stack, sp + 4, sv_2mortal(newSViv(first)));
    av_store(stack, sp + 5, sv_2mortal(newSViv(last)));
    RECOVERSTACK;
    TkpCallback(&listPtr->xScrollCallback, 4);
}

/*
 *----------------------------------------------------------------------
 *
 * ListboxComputeWidths --
 *
 *	This procedure is invoked to completely recompute width
 *	information used for displaying listboxes and for horizontal
 *	scrolling.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	If "fontChanged" is non-zero then the widths of the individual
 *	elements are all recomputed.  In addition, listPtr->maxWidth is
 *	recomputed.
 *
 *----------------------------------------------------------------------
 */

static void
ListboxComputeWidths(listPtr, fontChanged)
    Listbox *listPtr;		/* Listbox whose geometry is to be
				 * recomputed. */
    int fontChanged;		/* Non-zero means the font may have changed
				 * so per-element width information also
				 * has to be computed. */
{
    register Element *elPtr;
    int dummy;
    XCharStruct bbox;

    listPtr->xScrollUnit = XTextWidth(listPtr->fontPtr, "0", 1);
    listPtr->maxWidth = 0;
    for (elPtr = listPtr->elementPtr; elPtr != NULL; elPtr = elPtr->nextPtr) {
	if (fontChanged) {
	    XTextExtents(listPtr->fontPtr, elPtr->text, elPtr->textLength,
		    &dummy, &dummy, &dummy, &bbox);
	    elPtr->lBearing = bbox.lbearing;
	    elPtr->pixelWidth = bbox.rbearing - bbox.lbearing;
	}
	if (elPtr->pixelWidth > listPtr->maxWidth) {
	    listPtr->maxWidth = elPtr->pixelWidth;
	}
    }
}

void bootC_Listbox _((void))
{
    char *file = __FILE__;

    listboxClassRec.package = "Listbox";
    listboxClassRec.stash = Nullhv; /* filled in by init_widget */
    listboxClassRec.newwidgetproc = TkpListboxCmd;
    listboxClassRec.newix = 0;
    listboxClassRec.widgetcmdproc = TkpListboxWidgetCmd;
    listboxClassRec.methods = listboxMethods;
    listboxClassRec.filename = file;
    init_widget(&listboxClassRec);
}
